package com.bergerkiller.bukkit.nolagg.saving;

import java.lang.reflect.Field;
import java.util.HashSet;
import java.util.Set;
import java.util.logging.Level;

import org.bukkit.Bukkit;
import org.bukkit.craftbukkit.CraftServer;

import com.bergerkiller.bukkit.nolagg.ChunkHandler;
import com.bergerkiller.bukkit.nolagg.CustomChunkLoader;
import com.bergerkiller.bukkit.nolagg.NoLagg;

import net.minecraft.server.Chunk;
import net.minecraft.server.ChunkProviderServer;
import net.minecraft.server.IChunkLoader;
import net.minecraft.server.World;
import net.minecraft.server.WorldServer;

@SuppressWarnings("rawtypes")
public class ChunkSaving {

	private static DummyWorld dummyWorld;
	private static DummyChunk dummyChunk;
	private static Field chunkloader;
	private static Field ticklist;
	public static boolean enabled = false;
	
	public static CraftServer getServer() {
		return (CraftServer) Bukkit.getServer();
	}
			
	public static void init() {
		try {
            dummyWorld = new DummyWorld();
			chunkloader = ChunkProviderServer.class.getDeclaredField("e");
			ticklist = World.class.getDeclaredField("L");
			chunkloader.setAccessible(true);
			ticklist.setAccessible(true);
			dummyChunk = new DummyChunk(dummyWorld);
			enabled = true;
		} catch (Throwable t) {
			NoLagg.log(Level.SEVERE, "Failed to initialize async chunk saving implementation:");
			t.printStackTrace();
			enabled = false;
		}
	}
	public static void deinit() {
		enabled = false;
		dummyWorld = null;
		ticklist = null;
		chunkloader = null;
	}
	
	@SuppressWarnings("unchecked")
	public static Set getTickList(WorldServer from) {
		if (ticklist == null) return null;
		try {
			return new HashSet((Set) ticklist.get(from));
		} catch (Exception e) {
			NoLagg.log(Level.SEVERE, "Failed to obtain tick lists from world:");
			e.printStackTrace();
			ticklist = null;
			return new HashSet();
		}
	}
	public static void setDummyTickList(Set ticks) {
		if (ticklist == null) return;
		if (ticks == null) ticks = new HashSet();
		try {
			ticklist.set(dummyWorld, ticks);
		} catch (Exception e) {
			NoLagg.log(Level.SEVERE, "Failed to apply tick lists to dummy world:");
			e.printStackTrace();
			ticklist = null;
		}
	}
	
	public static enum SaveMode {FAILED, BUFFERED, NATIVE};
	
	public static SaveMode saveChunk(final Chunk chunk, final Set ticks) {
		return saveChunk(false, chunk, ticks);
	}
	public static SaveMode saveChunk(boolean cloned, final Chunk chunk, final Set ticks) {
		if (!enabled || chunk == null) return SaveMode.FAILED;
		try {
			//prepare dummy
			dummyWorld.setTime(chunk.world.getTime());
			setDummyTickList(ticks);
			//dummyWorld used for saving
			IChunkLoader loader = (IChunkLoader) chunkloader.get(((WorldServer) chunk.world).chunkProviderServer);
			if (cloned) {
				dummyChunk.load(chunk);
				if (loader instanceof CustomChunkLoader) {
					((CustomChunkLoader) loader).saveDirect(dummyWorld, dummyChunk);
					return SaveMode.BUFFERED;
				} else {
					loader.a(dummyWorld, dummyChunk);
					return SaveMode.NATIVE;
				}
			} else {
				if (loader instanceof CustomChunkLoader) {
					((CustomChunkLoader) loader).saveDirect(dummyWorld, chunk);
					return SaveMode.BUFFERED;
				} else {
					loader.a(dummyWorld, chunk);
					return SaveMode.NATIVE;
				}
			}
		} catch (Exception e) {
			NoLagg.log(Level.WARNING, "Failed to save chunk " + ChunkHandler.toString(chunk) + ":");
			e.printStackTrace();
		}
		//deinit dummy world
		setDummyTickList(null);
		return SaveMode.FAILED;
	}

}
